// NTSHELL2.C   Simple extendable command interpreter 
//              illustrating use of WIN32 process and 
//              thread creation API.
//
// Copyright 1992 Ray Duncan
// PC Magazine * Ziff Davis Publishing

#include "windows.h"
#include "stdio.h"
#include "process.h"
#include "stdlib.h"
#include "string.h"
#include "ntshell.h"

//
// This table associates "intrinsic" commands with the 
// routines that carry them out.
//
struct decodeCmd commands[] = {
    "CLS",   doCls,
    "DOS",   doDos,
    "EXIT",  doExit, 
    "KILL",  doKill, };

char comspec[MAXEXENAMESIZE];               // pathname of CMD.EXE

char msg1[] = "\nCan't run CMD.EXE.\n";     // misc. error messages
char msg2[] = "\nCreateProcess failed.\n";  
char msg3[] = "\nNo COMSPEC in environment.\n"; 
char msg4[] = "\nCreateThread failed.\n";  
char msg5[] = "\nTerminateProcess failed.\n";  

HANDLE hConIn;                              // console input handle
HANDLE hConOut;                             // console output handle


VOID main(int argc, char *argv[])
{
    char buffer[INPUTSIZE];                 // keyboard input buffer

    // open handles for console input and output
    hConIn =  CreateFile("CONIN$", GENERIC_READ | GENERIC_WRITE,
                         FILE_SHARE_READ | FILE_SHARE_WRITE, 
                         NULL, OPEN_EXISTING, 0, NULL);
    hConOut = CreateFile("CONOUT$", GENERIC_READ | GENERIC_WRITE,
                         FILE_SHARE_READ | FILE_SHARE_WRITE, 
                         NULL, OPEN_EXISTING, 0, NULL);

    // bail out if we can't get console I/O handles
    if((hConIn == (HANDLE) -1) || (hConOut == (HANDLE) -1)) 
        exit(TRUE);               

    getComspec(comspec);                    // get pathname for CMD.EXE

    while(TRUE)                             // main interpreter loop
    {
        getCommand(buffer);                 // get command from user
        if(strlen(buffer))                  // ignore empty lines
            if(!intrinsic(buffer))          // if intrinsic, run worker
               extrinsic(buffer);           // else assume external program 
    }                                       // & try to create new process
}

//
//  intrinsic() -- try and match user's command with intrinsic 
//  command table.  If a match is found, run the associated routine
//  and return TRUE, else return FALSE.
//
UINT intrinsic(char *input)
{
    int i, j;                               // scratch variables
    char token[64];                         // scratch buffer

    while(*input == '\x20') input++ ;       // discard leading blanks

    i = strcspn(input, " ;/,-");            // search for delimiter
    memset(token, 0, sizeof(token));        // zero out token buffer
    strncpy(token, input, i);               // copy first input token
    strupr(token);                          // fold token to upper case

    for(j = 0; j < dim(commands); j++)      // search command table
    {
        if(!strcmp(commands[j].name, token))
        {                                   // if match, run routine,
            (*commands[j].fxn)(&input[i]);  // passing the command tail
            return(TRUE);                   // and return true status
        }
    }
    return(FALSE);                          // no match, return false
}

//
//  extrinsic() -- process an unrecognized command by attempting
//  to create a child process, using the first token of the input
//  as the program name.  If the child process is created, also
//  create a supervisor thread that will wait for the child's 
//  termination and then display a message.
//
VOID extrinsic(char *input)
{
    STARTUPINFO startupinfo;                // used by CreateProcess
    PROCESS_INFORMATION processinfo;        // used by CreateProcess
    char temp[256];                         // scratch formatting buffer
    DWORD tid;                              // returned thread ID

    // initialize process startup information structure
    memset(&startupinfo, 0, sizeof(STARTUPINFO));   
    startupinfo.cb = sizeof(STARTUPINFO);           
    startupinfo.lpTitle = "Child of NTSHELL";

    // now attempt to launch the child process
    if(!CreateProcess(NULL,                 // image filename 
                      input,                // command line
                      NULL,                 // process security attributes
                      NULL,                 // thread security attributes
                      TRUE,                 // inherit handles
                      NORMAL_PRIORITY_CLASS |   
                      CREATE_NEW_CONSOLE,   // process creation flags
                      NULL,                 // use current environment
                      NULL,                 // use current directory
                      &startupinfo,         // startup information
                      &processinfo))        // receives process info
        putMsg(msg2);                       // error message if failed
    else
    {
        // launch was successful, display PID and start supervisor thread
        wsprintf(temp, "\nNew process handle: %d\n", processinfo.hProcess);
        putMsg(temp);
        if(!CreateThread(NULL,              // thread security attributes
                         0,                 // stack size = default
                         (LPTHREAD_START_ROUTINE) supervisor, // starting address
                         processinfo.hProcess,  // argument for thread
                         0,                 // flags: run immediately
                         &tid))             // receives thread ID
            putMsg(msg4);                   // error message if failed
    }
}

//
// supervisor() - this routine is run as a separate thread each time
// a new child process is launched.  It blocks on the child's process
// handle and displays a message when the child terminates.
//
DWORD supervisor(HANDLE hProcess)
{
    char temp[256];

    WaitForSingleObject(hProcess, INFINITE); // wait for child to terminate
    wsprintf(temp, "\n\nProcess %d died.\n", hProcess);
    putMsg(temp);                           // audit child's termination
    return(0);                              // implied ExitThread(0)
}

//
//  getCommand() - issue prompt, get user's command from console,
//  and make it ASCIIZ.
//
VOID getCommand(char *buffer)
{
    DWORD count;

    putMsg("\nntshell: ");                  // display prompt 
    ReadFile(hConIn, buffer, INPUTSIZE, &count, NULL);
    buffer[count-2] = 0;                    // clip off CR-LF
    strupr(buffer);                         // fold to upper case 
}

//
//  getComspec() -- get the full pathname for CMD.EXE
//  from the "COMSPEC=" variable in the environment.
//
VOID getComspec(char *buffer)
{
    strcpy(buffer, getenv("COMSPEC"));      // get COMSPEC variable
                                            // from environment
    if(!buffer[0])                      
    {                                       // if no such variable,
        putMsg(msg3);                       // display error message
        exit(1);                            // and exit
    }
}

//
// putMsg() -- display a string on the console.
//
VOID putMsg(char *string)
{
    DWORD count;

    WriteFile(hConOut, string, strlen(string), &count, NULL);
}

//
//  These are the subroutines for NTSHELL's intrinsic commands.
//

VOID doCls(char *input)                     // CLS command = 
{                                           // erase the window
    DWORD count;
    COORD coord;
    CONSOLE_SCREEN_BUFFER_INFO si;

    GetConsoleScreenBufferInfo(hConOut, &si); // get screen dimensions

    coord.X = 0;                            // buffer coordinates for
    coord.Y = 0;                            // start of blanking = (0,0)

    // fill screen with blanks, set cursor at upper left corner
    FillConsoleOutputCharacter(hConOut, ' ', si.dwSize.X * si.dwSize.Y, 
        coord, &count);
    SetConsoleCursorPosition(hConOut, coord);
}   

VOID doDos(char *input)                     // DOS command = start
{                                           // child copy of CMD.EXE 
    if(spawnlp(P_WAIT, comspec, comspec, NULL))
        putMsg(msg1);                       // error message if failed
}

VOID doExit(char *input)                    // EXIT command =
{                                           // terminate NTSHELL
    exit(FALSE);                            // with success status
}

VOID doKill(char *input)                    // KILL command = kill
{                                           // child process by handle
    HANDLE hProcess;                        

    while(*input == '\x20') input++ ;       // discard leading blanks
    hProcess = (HANDLE) atoi(input);        // convert handle number
    if(!TerminateProcess(hProcess, 0))      // terminate child process
        putMsg(msg5);                       // error message if failed
}


